from abc import ABC
from typing import List, Any, Union, Optional
from see137.common import InjectableObject
from see137.common import ComponentAbstract
from loguru import logger




class Pipeline(ABC, InjectableObject):
    """ 
        Use pipeline to run a large sequence of commands back to back. 
        Results from the previous result to process the next component/or pipeline.
    """
    def __init__(self, steps: List[Union['ComponentAbstract', 'Pipeline']]=[], name:Optional[str]=None, processor=None):
        self.steps:List[Union['ComponentAbstract', 'Pipeline']] = steps
        self.history = []
        self.name = name or self.__class__.__name__
        self._within_pipeline = []
        if processor is not None:
            self.processor = processor


    @property
    def within_pipeline(self):
        return self._within_pipeline
    
    @within_pipeline.setter
    def within_pipeline(self, _name):
        self._within_pipeline.append(_name)

    def verify(self):
        """Verify Steps
            Verify if the steps are all component abstracts and that it's not empty
        """
        if not bool(self.steps):
            raise AttributeError("Steps are Falsy (likely haven't added or replaced the variable). Please add a ComponentAbstract")
        for step in self.steps:
            if not isinstance(step, ComponentAbstract) and not isinstance(step, Pipeline):
                raise TypeError(
                    "One of the steps isn't an instance of ComponentAbstract or Pipeline. Only those two types."
                )

    @property
    def is_history(self):
        if len(self.history) > 0:
            return True
        return False

    @property
    def latest_history(self):
        if self.is_history:
            return self.history[-1]
        return None

    def pipe(self, component: Union[Union['ComponentAbstract', 'Pipeline'], List[Union['ComponentAbstract', 'Pipeline']]]):
        """Add component to pipeline. Use instead of the constructure.

        Arguments:
            item {Union[ComponentAbstract, Pipeline]} -- [description]
        """
        if isinstance(component, list):
            self.steps.extend(component)
            return self
        else:
            self.steps.append(component)
            return self

    @logger.catch(reraise=True)
    def step(self, item:Any, override=False, is_loud=True, **kwargs):
        """Runs through a sequence of events. Takes the previous one and passes it onto the next one. Done to monitor highly relational queries.

        Arguments:
            item {Any} -- The item we itend to process. For many components this is an addict dictionary. 
            
            NOTE: In future should have more constrains for this parameter.

        Returns:
            [Any] -- Returns the last item in pipeline. That can either be a pipeline response, or the response of a component.
        """
        self.verify()
        self.reset()
        for index, _step in enumerate(self.steps):
            step_result = None
            if is_loud:
                logger.debug(f"Running {_step.name}")
            _step.within_pipeline = self.name
            _step.processor = self.processor
            
            if index == 0:
                step_result = _step.step(item, override=override, is_loud=is_loud, **kwargs)
            if self.is_history:
                step_result = _step.step(self.latest_history, override=override, is_loud=is_loud, **kwargs)
            self.history.append(step_result)
        return self.latest_history

    def reset(self):
        self.history = []





if __name__ == "__main__":

    pipe = Pipeline([
        "find_item_metadata", "pull_data_from_outside", "store_data",
        "find_strategy", "find_send_location", "create_model", "send_model"
    ])
    pipe.step("sample_id")


from see137.parameters import Parameter
from see137.parameters.fields import CurrencyFields, ExchangeFields
from jamboree.utils.core import consistent_hash

class PortfolioParameter(Parameter):
    def __init__(self, currency:CurrencyFields, exchange:ExchangeFields, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.currency_fields = currency
        self.exchange_fields = exchange
        # We'd split this between currency, exchange details, and risk details
    
    @property
    def initial(self):
        return self.currency_fields.initial
    
    @property
    def base(self):
        return self.currency_fields.base
    
    @property
    def exchange(self):
        return self.exchange_fields.name
    
    @property
    def slippage(self):
        return self.exchange_fields.slippage



import random
import time
import uuid
from pprint import pprint
from typing import Any, List, Optional

import maya
import pandas as pd
from addict import Dict
from jamboree import Jamboree
from jamboree.handlers.complex.engines import FileEngine
# from jamboree.middleware.procedures import (ModelProcedureAbstract,
#                                             ModelProcedureManagement,
#                                             ProcedureAbstract,
#                                             ProcedureManagement)
from jamboree.utils.support.search import querying
from loguru import logger

# logger.disable(__name__)

class StrategyEngine(FileEngine):
    """ 
        Save and load strategies dynamically by metadata and ID.

        Use to make decisions on multiple assets. 

        Uses context to make decisions and provide some clean up during the process :
        ```py
            with strategy as strat:
                strat.step(dataframe)
        ```
    """

    def __init__(self, processor=None, **kwargs):
        super().__init__(processor=processor, **kwargs)
        self.entity = "deployed_strategies"
        self.current_strategy = None
        

    def init_specialized(self, **kwargs):
        super().init_specialized(**kwargs)

    def verify(self):
        """ Disabling verify. """
        pass

    def open_context(self):
        if not self.file_reset:
            self.reset()
        

    def close_context(self):
        pass


    def enterable(self):
        """ Return the object we want to enter into """
        return self.strategy

    def custom_post_load(self, item):
        self.current_strategy = item

    @property
    def strategy(self):
        if self.current_strategy is not None:
            return self.current_strategy
        raise AttributeError("No strategy set and found")
    
    def file_from_dict(self, item:Dict):
        reloaded = StrategyEngine(
            processor=self.processor,
            name=item.name,
            category=item.category,
            subcategories=item.subcategories,
            submetatype=item.submetatype,
            abbreviation=item.abbreviation
        )
        return reloaded

    def reset(self):
        super().reset()
        if self.file_reset == True and self.current_strategy is None:
            if self.blobfile is not None:
                self.current_strategy = self.blobfile

class SampleStrategy(object):
    def step(self, df:pd.DataFrame):
        """ Preprocess the dataframe and make decisions using the step function. """
        return random.choice([0, 1, 2])

@logger.catch
def main_strategy():
    jam_processor = Jamboree()
    strat = SampleStrategy()
    strat_engine = StrategyEngine(
        processor=jam_processor,
        name="SAMPLE_NAME",
        category="SAMCATSASS",
        subcategories={"sam": "ple"},
        submetatype="OTHER SMAPLE",
        abbreviation="SAP",
        blobfile=strat
    )
    strat_engine.reset()
    while True:
        # We'd `pick` a strategy by `episode`. We'd first get an episode, get the strategy, then get the strategy, then pass in the data 
        strategy = strat_engine.first(
            name="SAMPLE_NAME",
            category="SAMCATSASS",
            submetatype="OTHER SMAPLE"
        )
        with strategy as stray:
            logger.warning(stray.step(1223))

if __name__ == "__main__":
    main_strategy()



import time as _time
import math
import numpy
import random
from see137.stochastic.generators.geometric_brownian import geometric_brownian_motion_log_returns
from see137.stochastic import ModelParameters, convert_to_prices



def jump_diffusion_process(param):
    """
    This method produces a sequence of Jump Sizes which represent a jump diffusion process. These jumps are combined
    with a geometric brownian motion (log returns) to produce the Merton model.
    :param param: the model parameters object
    :return: jump sizes for each point in time (mostly zeroes if jumps are infrequent)
    """
    assert isinstance(param, ModelParameters)
    s_n = time = 0
    small_lamda = -(1.0 / param.lamda)
    jump_sizes = []
    for k in range(0, param.all_time):
        jump_sizes.append(0.0)
    
    last_step = 0.0
    while s_n < param.all_time:
        s_n += small_lamda * math.log(random.uniform(0, 1))
        last_step = _time.time()
        for j in range(0, param.all_time):
            delta = _time.time()-last_step
            if delta > 1:
                print("lingering gere")
            if time * param.all_delta <= s_n * param.all_delta <= (j + 1) * param.all_delta:
                jump_sizes[j] += random.normalvariate(param.jumps_mu, param.jumps_sigma)
                break
            # print()
        time += 1
    return jump_sizes


def geometric_brownian_motion_jump_diffusion_log_returns(param):
    """
    This method constructs combines a geometric brownian motion process (log returns) with a jump diffusion process
    (log returns) to produce a sequence of gbm jump returns.
    :param param: model parameters object
    :return: returns a GBM process with jumps in it
    """
    assert isinstance(param, ModelParameters)
    jump_diffusion = jump_diffusion_process(param)
    geometric_brownian_motion = geometric_brownian_motion_log_returns(param)
    return numpy.add(jump_diffusion, geometric_brownian_motion)


def geometric_brownian_motion_jump_diffusion_levels(param):
    """
    This method converts a sequence of gbm jmp returns into a price sequence which evolves according to a geometric
    brownian motion but can contain jumps at any point in time.
    :param param: model parameters object
    :return: the price levels
    """
    return convert_to_prices(param, geometric_brownian_motion_jump_diffusion_log_returns(param))


from see137.utils.trades import TradeType

class TradePercentage(object):
    """A trade object for use within trading environments."""

    def __init__(self, symbol: str, trade_type: 'TradeType', percentage=0.4):
        """
        Arguments:
            symbol: The exchange symbol of the instrument in the trade (AAPL, ETH/USD, NQ1!, etc).
            trade_type: The type of trade executed (0 = HOLD, 1=LIMIT_BUY, 2=MARKET_BUY, 3=LIMIT_SELL, 4=MARKET_SELL).
            amount: The amount of the instrument in the trade (shares, satoshis, contracts, etc).
            price: The price paid per instrument in terms of the base instrument (e.g. 10000 represents $10,000.00 if the `base_instrument` is "USD").
        """
        self._symbol = symbol
        self._trade_type = trade_type
        self._percentage = percentage
        self._amount = 0.0
        self._price = 0.0

    def copy(self) -> 'TradePercentage':
        """Return a copy of the current trade object."""
        return TradePercentage(symbol=self._symbol, trade_type=self._trade_type, percentage=self._percentage)

    
    @property
    def symbol(self) -> str:
        """The exchange symbol of the instrument in the trade (AAPL, ETH/USD, NQ1!, etc)."""
        return self._symbol

    @symbol.setter
    def symbol(self, symbol: str):
        self._symbol = symbol

    @property
    def trade_type(self) -> 'TradeType':
        """The type of trade ("buy", "sell", "hold", etc)."""
        return self._trade_type

    @trade_type.setter
    def trade_type(self, trade_type: 'TradeType'):
        self._trade_type = trade_type

    @property
    def amount(self) -> float:
        """The amount of the instrument in the trade (shares, satoshis, contracts, etc)."""
        return self._amount

    @amount.setter
    def amount(self, amount: float):
        self._amount = amount


    @property
    def is_hold(self) -> bool:
        """
        Returns:
            Whether the trade type is non-existent (i.e. hold).
        """
        return self._trade_type.is_hold

    @property
    def is_buy(self) -> bool:
        """
        Returns:
            Whether the trade type is a buy offer.
        """
        return self._trade_type.is_buy

    @property
    def is_sell(self) -> bool:
        """
        Returns:
            Whether the trade type is a sell offer.
        """
        return self._trade_type.is_sell


"""
geometric_brownian_motion_levels
heston_model_levels
geometric_brownian_motion_jump_diffusion_levels
cox_ingersoll_ross_levels
ornstein_uhlenbeck_levels
"""

from .geometric_brownian import geometric_brownian_motion_levels
from .heston import heston_model_levels
from .jump_diffusion import geometric_brownian_motion_jump_diffusion_levels
from .cox import cox_ingersoll_ross_levels
from .ornstien import ornstein_uhlenbeck_levels


from abc import ABC
import uuid
"""
    Configuration Abstract
"""

# NOTE: We don't need this. It's more of a code smell. Creating objects using a pipeline instead. 

class Parameter(ABC):
    def __init__(self, *args, **kwargs):
        pass


import random
import uuid
import time
import maya
import pandas as pd
import requests
from crayons import cyan, magenta
from jamboree import Jamboree
from loguru import logger

from see137.config.fields import StepField
from see137.handlers.strategy.core import StrategyEngine
from see137.search.backtesting import BacktestManager


class SampleStrategy(object):
    def step(self, df: pd.DataFrame):
        return random.choice([0, 1, 2])


class ProblemSetup:
    def __init__(self):
        self.name = "Test Strategy"
        self.description = "Test Description"
        self.user_id = uuid.uuid4().hex
        self.processor = Jamboree()
        # self.must_have = ['name', 'description', 'userid']
        self.url = "http://localhost:6066"
        self.manager = BacktestManager()
        self.manager.processor = self.processor
        self.backtest_id: str = ""
        self.strategy_id: str = ""
        self.strategy = StrategyEngine(processor=self.processor)
        self.cash = 10000.0
        """ 
            name="SAMPLE_NAME",
            category="SAMCATSASS",
            subcategories={"sam": "ple"},
            submetatype="OTHER SMAPLE",
            abbreviation="SAP",
        """

    def create_backtest(self):
        self.backtest_id = self.manager.create.backtest.init(
            name=self.name, description=self.description, userid=self.user_id)
        # self.current_strategy_engine
        logger.info(
            f"Create backtest config in DB: {magenta(self.backtest_id)}")
    def save_strategy_getmeata(self):
        """ Save strategy and get metadata"""
        self.strategy['name'] = "SAMPLE_NAME"
        self.strategy['subcategories'] = {"sam": "ple"}
        self.strategy['category'] = "SAMCATSASS"
        self.strategy['submetatype'] = "OTHER SMAPLE"
        self.strategy['abbreviation'] = "SAP"
        self.strategy.blobfile = SampleStrategy()
        self.strategy.reset()
        self.strategy_id = str(self.strategy.metaid)

    def update_strategyid(self):
        self.manager.update.backtest.UpdateID(self.backtest_id,
                                              strategy=self.strategy_id)
        btest = self.manager.info.backtest.FindById(self.backtest_id)

        vals = list(btest.values())
        logger.warning(self.strategy_id)
        logger.info(
            f"Saving strategy and getting the metadata id associated with it {cyan(btest['name'], bold=True)}"
        )
        logger.error(random.choice(vals))
        current_strategy = self.strategy.pick(btest['strategy'])
        with current_strategy as srt:
            logger.success(f"Strategy action will be: {srt.step(123)}")

    def setup_time(self):
        start_time = maya.when("2007-05-02")
        end_time = maya.when("2020-04-19")
        step_information = StepField()
        step_information.hours = 4
        logger.info(start_time)
        logger.success(end_time)
        logger.warning(step_information.to_epoch())

        self.manager.update.backtest.UpdateID(
            self.backtest_id,
            start_epoch=start_time._epoch,
            end_epoch=end_time._epoch,
            step_size=step_information.to_epoch()
        )

    def setup_portfolio(self):
        self.manager.update.backtest.UpdateID(self.backtest_id,
                                              exchange="binance", start_cash=self.cash)
        btest = self.manager.info.backtest.FindById(self.backtest_id)
        logger.warning(type(btest['exchange']))

    def create_universe(self):
        """ Moving this to another script file"""
        pass

    def add_assets_to_universe(self):
        """ Moving this to another script file. """
        pass

    def add_universe(self):
        """ Add a universe of assets """
        logger.info("add a universe of assets")
        self.manager.update.backtest.UpdateID(self.backtest_id,
                                              universe=uuid.uuid4().hex)
        btest = self.manager.info.backtest.FindById(self.backtest_id)
        logger.warning(btest)

    def send_start_command(self):
        response = requests.post(f"{self.url}/episodes/init",
                                 json={"episode_number": 2, "backtest_id": self.backtest_id})
        logger.info(response.json())

    def start(self):
        self.create_backtest()
        self.save_strategy_getmeata()
        self.update_strategyid()
        self.setup_time()
        self.setup_portfolio()
        self.add_universe()
        self.send_start_command()
        time.sleep(0.01)

# Send an HTTP Request to Faust and work backwards.

# 1. Send HTTP request
# 2. Configure searchable item for backtest
# 3. Pull searchable item from HTTP request
# 4. Create an episode upon request from searchable item pulled
# 5. 
if __name__ == "__main__":
    ProblemSetup().start()



import time
import maya
import datetime as dateeee
from maya import MayaDT
from patman import Patman
from typing import Any

class TiingoAPI(Patman):
    def __init__(self, host:str, token=""):
        super().__init__(host)
        self._token = token
        self._period = ""
        self._freq = "1min"
    
    @property
    def token(self):
        return self._token
    
    @token.setter
    def token(self, _token):
        self._token = _token
    
    @property
    def freq(self):
        return self._freq
    
    @freq.setter
    def freq(self, _freq):
        self._freq = _freq
    
    def realtime_ticker(self, ticker:str):
        url = f"iex?tickers={ticker}&token={self.token}"
        call = self._create_url_call(url)
        return self.get(url)
    
    def standardize_date(self, time_data:Any):
        standardized_date = None
        if isinstance(time_data, time.struct_time):
            standardized_date = maya.MayaDT.from_struct(time_data)
        elif isinstance(time_data, str):
            standardized_date = maya.when(time_data)
        elif isinstance(time_data, dateeee.datetime):
            standardized_date = maya.MayaDT.from_datetime(time_data)
        elif isinstance(time_data, float):
            standardized_date = maya.MayaDT(time_data)
        return standardized_date
    
    def _get_date_string(self, _date:maya.MayaDT):
        return f"{_date.year}-{_date.month}-{_date.day}"
    
    def _aggregate_standard(self, _time:Any):
        current_time = self.standardize_date(_time)
        return self._get_date_string(current_time)
    
    def get_intraday(self, ticker, start_date, end_date=time.time()):
        start_date = self._aggregate_standard(start_date)
        end_date = self._aggregate_standard(end_date)
        url = f"iex/{ticker}/prices?startDate={start_date}&endDate={end_date}&resampleFreq={self.freq}&token={self.token}"
        # print(url)
        # self.get(url)
        return {}
    
    def get_daily(self, ticker:str, start_date:Any=(time.time()-(3600 * 280000)), end_date:Any=time.time()):
#         print(start_date)
        start_date = self._aggregate_standard(start_date)
        end_date = self._aggregate_standard(end_date)
        url = f"tiingo/daily/{ticker}/prices?startDate={start_date}&endDate={end_date}&token={self.token}"
        return self.get(url)

if __name__ == "__main__":
    tingo = TiingoAPI("api.tiingo.com")
    tingo['protocol'] = "https"
    tingo.token = "f6deea4df3f704452993854b3d30e3ac4af6560b"
    print(tingo.realtime_ticker("GE"))



# from .data_handler import DataHandler

# from .base_handlers.main import DataSet
# from .base_handlers.runnable import RunnableSet
# from .base_handlers.simulated import SimulatedData
# from .base_handlers.backtest import BacktestDataset
# from .base_handlers.price import PriceSet
# from .base_handlers.live import LiveSet

from .stochastic_generator import PriceGenerator


import uuid
import pandas as pd
import numpy as np
import maya
from jamboree import Jamboree, DBHandler
from jamboree.handlers.default import TimeHandler

from loguru import logger
from typing import List, Dict, Any, Optional
from toolz.itertoolz import pluck

from pypfopt.expected_returns import mean_historical_return
from pypfopt.risk_models import CovarianceShrinkage
from pypfopt.efficient_frontier import EfficientFrontier

logger.disable(__name__)
"""
    Note: Will remove soon. Going through a concept.
"""



class AllocationHandler(DBHandler):
    """Abstract handler that we use to keep track of information.
    """

    def __init__(self, limit=500, spacing=5, unit="hours"):
        super().__init__()
        self.entity = "allocation"
        self.required = {
            "user_id": str,
            "exchange": str,
            "episode": str,
            "live": bool
        }
        self._balance = 0
        self._limit = limit
        self._spacing = spacing
        self._unit = unit
        self._time: Optional[TimeHandler] = None

    # -------------------------------------------------------------------
    # --------------------- Properties & Setters ------------------------
    # -------------------------------------------------------------------

    @property
    def limit(self):
        return self._limit

    @limit.setter
    def limit(self, limit):
        self._limit = limit

    @property
    def spacing(self):
        return self._spacing

    @spacing.setter
    def spacing(self, space):
        self._spacing = space

    @property
    def time(self):
        if self._time is None:
            raise AttributeError("Time not found")
        return self._time

    @time.setter
    def time(self, _time: TimeHandler):
        self._time = _time

    """
        Notes:
            TODO: Change how the units are done.
    """

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, _unit):
        self._unit = _unit

    @property
    def assets(self) -> list:
        """ Get all of the assets that we're watching for the exchange """
        return self.latest_asset().get("assets", [])

    @property
    def allocation(self):
        return self.latest_allocation().get("allocation", {})

    @property
    def current_step(self):
        alt = {"detail": "steps"}
        count = self.count(alt=alt)
        return count

    @property
    def last_time(self):
        _late = self.latest_timestamp()
        return _late.get("epoch", maya.now()._epoch)

    @property
    def should_rebalance(self) -> bool:
        """ Depending on the scenario, we determine if we should rebalance."""
        """
            if live:
                check if the determined amount of time has passed since we've rebalanced
            else:
                check if the predetermined number of steps have happened.
        """
        if self["live"]:
            current_epoch = maya.now()._epoch
            expected = maya.MayaDT(self.last_time).add(
                **{f"{self.unit}": self.spacing})._epoch
            if current_epoch >= expected:
                return True
        else:
            if self.current_step % self.spacing == 0:
                return True
        return False

    def latest_asset(self):
        alt = {"detail": "assets"}
        asset = self.last(alt=alt)
        return asset

    def latest_allocation(self):
        alt = {"detail": "allocation"}
        asset = self.last(alt=alt)
        return asset

    def latest_timestamp(self):
        alt = {"detail": "timestamp"}
        asset = self.last(alt=alt)
        return asset

    def timestamp_count(self) -> int:
        """ Get the count of the asset adds we have. Not the number of assets. """
        alt = {"detail": "timestamp"}
        count = self.count(alt=alt)
        return count

    def asset_count(self) -> int:
        """ Get the count of the asset adds we have. Not the number of assets. """
        alt = {"detail": "assets"}
        count = self.count(alt=alt)
        return count

    def count_steps(self):
        alt = {"detail": "steps"}
        count = self.count(alt=alt)
        return count

    def save_allocation(self, allocation: dict):
        """ Save monitored assets """
        alt = {"detail": "allocation"}
        monitored = {
            "allocation": allocation,
            "time": self.time.head
        }
        self.save(monitored, alt=alt)

    def save_timestamp(self, _time=None):
        """ Save monitored assets """
        if _time is None:
            _time = maya.now()._epoch
        alt = {"detail": "timestamp"}
        monitored = {
            "epoch": _time
        }
        self.save(monitored, alt=alt)

    def save_step(self):
        alt = {"detail": "steps"}
        monitored = {}
        self.save(monitored, alt=alt)

    def _filter_assets(self, all_assets: dict) -> dict:
        assets = self.assets
        users_portfolio_to_rebalance = {}
        _all_assets = all_assets.get("assets", None)
        if _all_assets is None:
            return {}
        for asset in assets:
            current_asset = _all_assets.get(asset, None)
            if current_asset is not None:
                users_portfolio_to_rebalance[asset] = current_asset

        return users_portfolio_to_rebalance

    def _extract_frame(self, portfolio_assets: dict):
        time_is_filled = False
        time_index = []
        portfolio_close = {}
        for asset in portfolio_assets.keys():
            current_pricing = portfolio_assets.get(asset, [])
            portfolio_close[asset] = list(pluck("close", current_pricing))
            if time_is_filled == False:
                time_index = list(pluck("time", current_pricing))
                time_is_filled = True

        pandas_index = pd.to_datetime(time_index, unit='s')
        pclose = pd.DataFrame(portfolio_close, index=pandas_index)
        return pclose

    def _finish_step(self, is_rebalance=False):
        """ Finish update that we're not """
        if is_rebalance == True:
            # Give a timestamp of the last rebalance.
            self.save_timestamp()
        self.save_step()
    def step(self, rebal_frame: pd.DataFrame) -> None:
        """ Accept all of the assets for a given allocation and begin"""

        # logger.info(rebal_frame)
        rebal_len = len(rebal_frame)
        if rebal_len == 0:
            # Don't rebalance any further
            self._finish_step()
            return

        if self.should_rebalance == True:
            # Set timestamp aka, the last time we did a rebalance
            mu = mean_historical_return(rebal_frame)
            S = CovarianceShrinkage(rebal_frame).ledoit_wolf()
            ef = EfficientFrontier(mu, S)
            weights = ef.max_sharpe()
            logger.info((S, weights))
            current_allocation = ef.clean_weights()
            converted_allocation = {}
            for k, v in current_allocation.items():
                if isinstance(v, np.generic):
                    converted_allocation[k] = v.item()
                    continue
                converted_allocation[k] = v
            self.save_allocation(converted_allocation)
            self._finish_step(is_rebalance=True)

        self._finish_step()

    def new_step_test(self, rebal_frame: pd.DataFrame) -> None:
        """ Accept all of the assets for a given allocation and begin"""

        rebal_len = len(rebal_frame)
        if rebal_len == 0:
            # Don't rebalance any further
            self._finish_step()
            return

        if self.should_rebalance == True:
            # Set timestamp aka, the last time we did a rebalance
            mu = mean_historical_return(rebal_frame)
            S = CovarianceShrinkage(rebal_frame).ledoit_wolf()
            ef = EfficientFrontier(mu, S)
            weights = ef.max_sharpe()
            logger.info(weights)
            current_allocation = ef.clean_weights()
            self.save_allocation(current_allocation)
            self._finish_step(is_rebalance=True)

        self._finish_step()
    def rebalance(self, rebalance_dataframe: pd.DataFrame):
        logger.info(rebalance_dataframe)

    def _reset_timestamp(self):
        if self.timestamp_count() == 0:
            self.save_timestamp(_time=maya.when('12 years ago')._epoch)

    def reset(self):
        self._reset_timestamp()


def experiment_main():
    jam = Jamboree()


if __name__ == "__main__":
    experiment_main()



from .accessor import SingleAccessor
from .creation import SimulationCreator
from .injectable import InjectableObject
from .component import ComponentAbstract
from .pipeline import Pipeline
from .addict_comp import AddictComponent
from .addict_manager import AddictWithManagerComponent
from .addict_pipeline import AddictPipeline
from .addict_pipeline import AddictPipeline as PipelineA


import numpy
from see137.stochastic import brownian_motion_log_returns

def ornstein_uhlenbeck_levels(param):
    """
    This method returns the rate levels of a mean-reverting ornstein uhlenbeck process.
    :param param: the model parameters object
    :return: the interest rate levels for the Ornstein Uhlenbeck process
    """
    ou_levels = [param.all_r0]
    brownian_motion_returns = brownian_motion_log_returns(param)
    for i in range(1, param.all_time):
        drift = param.ou_a * (param.ou_mu - ou_levels[i-1]) * param.all_delta
        randomness = brownian_motion_returns[i - 1]
        ou_levels.append(ou_levels[i - 1] + drift + randomness)
    return numpy.array(ou_levels)


import numpy as np

from see137.utils.slippage import SlippageModel
from see137.utils.trades import Trade, TradeType


class RandomUniformSlippageModel(SlippageModel):
    """A uniform random slippage model."""

    def __init__(self, max_price_slippage_percent: float = 3.0, max_amount_slippage_percent: float = 0.0):
        """
        Arguments:
            max_price_slippage_percent: The maximum random slippage to be applied to the fill price. Defaults to 3.0 (i.e. 3%).
            max_amount_slippage_percent: The maximum random slippage to be applied to the fill amount. Defaults to 0.
        """
        self.max_price_slippage_percent = max_price_slippage_percent
        self.max_amount_slippage_percent = max_amount_slippage_percent

    def fill_order(self, trade: Trade, current_price: float) -> Trade:
        amount_slippage = np.random.uniform(0, self.max_amount_slippage_percent / 100)
        price_slippage = np.random.uniform(0, self.max_price_slippage_percent / 100)

        fill_amount = trade.amount * (1 - amount_slippage)
        fill_price = current_price
            
        if trade.trade_type is TradeType.MARKET_BUY:
            fill_price = current_price * (1 + price_slippage)
        elif trade.trade_type is TradeType.LIMIT_BUY:
            fill_price = max(current_price * (1 + price_slippage), 1e-3)
            if fill_price > trade.price:
                fill_price = trade.price
                if fill_price == 0:
                    fill_price = 0.00000001
                fill_amount *= trade.price / fill_price

        elif trade.trade_type is TradeType.MARKET_SELL:
            fill_price = current_price * (1 - price_slippage)
        elif trade.trade_type is TradeType.LIMIT_SELL:
            fill_price = max(current_price * (1 - price_slippage), 1e-3)
            if fill_price < trade.price:
                fill_price = trade.price
                
                fill_amount *= fill_price / trade.price

        return Trade(trade.symbol, trade.trade_type, amount=fill_amount, price=fill_price)



# As a user, I'd like to pull the attached process strategy, pull the data, process the data, get a decision from that data
from addict import Dict as ADict
from typing import List
from jamboree.handlers.default.multi import MultiDataManagement
from see137.handlers.strategy.core import StrategyEngine
from see137.common import Pipeline, ComponentAbstract, AddictWithManagerComponent
from loguru import logger

class PullUniverse(AddictWithManagerComponent):
    """
        Pulls all of the universe asset_ids
    """
    def __init__(self):
        super(PullUniverse, self).__init__()
        self.required_settings = {}

        # Backtest configuration dict. Can be saved into the database immediately.
        
        self.required_items = {
            "config": dict
        }

        self.__private_multi = MultiDataManagement()
        self.mult_handler = None
    
    def check_prev(self, item):
        return super().check_prev(item)

    def check_all(self, item):
        return super().check_all(item)
    
    def check_item(self, item):
        return super().check_item(item)
    
    @property
    def multi(self):
        self.__private_multi.processor = self.processor
        return self.__private_multi

    def check_multi(self, universe:str):
        """Check the if the multi-data handler exists. Should return none if not.  

        Arguments:
            universe {str} -- [description]

        Returns:
            bool -- A boolean representing if a multi-data handler exists.
        """
        self.mult_handler = self.multi.pick(universe)
        if not bool(self.mult_handler):
            # Would return true even if
            return False
        return True
    
    def validate_asset_num(self):
        ids = self.mult_handler.source_ids
        if len(ids) > 0:
            return True
        return False

    def process(self, item:ADict):
        """Given an id inside of the item dictionary `backtest`, get information using that id

            We're expecting a dynamic `universe` key inside of the dictionary.

        Arguments:
            item {ADict} -- Addict modified dictionary. Saved according to backtest data.

        Returns:
            ADict -- Modified dictionary with action and percentage inside of it
        """
        universe_id:str = item.config.universe
        if not self.settings.override:
            is_multi = self.check_multi(universe_id)
            if is_multi:
                if self.validate_asset_num():
                    item.valid_multi=True
                    item.failed = False
                    item.msg = "Successfully confirmed the Universe is valid"
                    item.asset_metadata = self.mult_handler.source_ids
                    return item
            item.valid_multi=False
            item.failed = True
            item.msg = "Universe was not valid"
            item.asset_metadata = []
            return item
        
        item.valid_multi=True
        item.msg = "Universe was not valid"
        item.asset_metadata = []
        return item

    # @logger.catch(reraise=True)
    def step(self, item:ADict, **settings):
        return super().step(item, **settings)


import math
import numpy
from see137.stochastic import ModelParameters, convert_to_prices, brownian_motion_log_returns


def geometric_brownian_motion_log_returns(param):
    """
    This method constructs a sequence of log returns which, when exponentiated, produce a random Geometric Brownian
    Motion (GBM). GBM is the stochastic process underlying the Black Scholes options pricing formula.
    :param param: model parameters object
    :return: returns the log returns of a geometric brownian motion process
    """
    assert isinstance(param, ModelParameters)
    wiener_process = numpy.array(brownian_motion_log_returns(param))
    sigma_pow_mu_delta = (param.gbm_mu - 0.5 * math.pow(param.all_sigma, 2.0)) * param.all_delta
    return wiener_process + sigma_pow_mu_delta


def geometric_brownian_motion_levels(param):
    """
    Returns a sequence of price levels for an asset which evolves according to a geometric brownian motion
    :param param: model parameters object
    :return: the price levels for the asset
    """
    return convert_to_prices(param, geometric_brownian_motion_log_returns(param))


# As a user, I'd like to pull the attached process strategy, pull the data, process the data, get a decision from that data
from addict import Dict as ADict
from typing import List
from jamboree.handlers.default.multi import MultiDataManagement
from see137.handlers.strategy.core import StrategyEngine
from see137.common import Pipeline, ComponentAbstract, AddictWithManagerComponent
from loguru import logger

class CheckStrategy(AddictWithManagerComponent):
    """
        Checks to see if strategy is valid

        
    """
    def __init__(self):
        super(CheckStrategy, self).__init__()

        # Backtest configuration dict. Can be saved into the database immediately.
        

        self.required_items = {
            "config": dict
        }
        
        
        self.__private_strategy = StrategyEngine()
    
    def check_prev(self, item):
        return super().check_prev(item)

    def check_all(self, item):
        return super().check_all(item)
    
    def check_item(self, item):
        return super().check_item(item)

    @property
    def strategy(self):
        self.__private_strategy.processor = self.processor
        return self.__private_strategy

    def check_strat(self, universe:str):
        """Check the if the multi-data handler exists. Should return none if not.  

        Arguments:
            universe {str} -- [description]

        Returns:
            bool -- A boolean representing if a multi-data handler exists.
        """
        try:
            self.strat_handler = self.strategy.pick(universe)
        except AttributeError:
            return False
        return True
    
    def validate_strategy(self):
        """Checks if the stategy blob is loaded into memory

        Returns:
            [type] -- [description]
        """
        if self.strat_handler.current_strategy is not None:
            return True
        return False
    


    def process(self, item:ADict):
        """Given an id inside of the item dictionary `backtest`, get information using that id

            We're expecting a dynamic `universe` key inside of the dictionary.

        Arguments:
            item {ADict} -- Addict modified dictionary. Saved according to backtest data.

        Returns:
            ADict -- Modified dictionary with action and percentage inside of it
        """

        strategy:str = item.config.strategy
        override = self.settings.override
        
        if override:
            item.valid_strategy=True
            item.msg = "Skipping strategy check due to override"
            return item
        is_strat = self.check_strat(strategy)
        if is_strat:
            if self.validate_strategy():
                item.valid_strategy=True
                item.failed = False
                item.msg = "Successfully confirmed the strategy is valid"
                return item
        item.valid_strategy=False
        item.failed = True
        item.msg = "Strategy didn't meet criteria [existing, having assets within it]"
        return item
        

    
    def step(self, item:ADict, **settings):
        return super().step(item, **settings)


class ExchangeFields:
    

    def __init__(self, name, broker_fee=0.005, slippage=0.005):
        self.name:str=name
        self.broker_fee:float=broker_fee
        self.slippage:float=slippage
        self.verify()
    
    def verify(self):
        if not isinstance(self.name, str):
            raise TypeError("Exchange name needs to be a string")
        

        splintered = "".join(self.name.split(" "))
        if not bool(splintered):
            raise ValueError("Name cannot be empty")
        


class CurrencyFields:
    def __init__(self, base:str="USD", initial:float=100000.0):
        self.base = base
        self.initial = initial
        self.verify()
    
    def verify(self):
        if self.initial <= 0:
            raise ValueError("Starting amount cannot be zero or negative")
        if not isinstance(self.base, str):
            raise TypeError("Base currency needs to be a string")


from abc import abstractmethod, ABC


from see137.utils.trades import Trade


class SlippageModel(ABC):
    """A model for simulating slippage on an exchange trade."""

    registered_name = "slippage"

    def __init__(self):
        pass

    @abstractmethod
    def fill_order(self, trade: Trade, **kwargs) -> 'Trade':
        """Simulate slippage on a trade ordered on a specific exchange.
        Arguments:
            trade: The trade executed on the exchange.
            **kwargs: Any other arguments necessary for the model.
        Returns:
            A filled `Trade` with the `price` and `amount` adjusted for slippage.
        """

        raise NotImplementedError()



import copy
import crayons as cy
from loguru import logger

# logger.add(sys.stdout, format="{time} - {level} - {message}")
def symbol_split(symbol):
    if not isinstance(symbol, str):
        return None
    
    s = symbol.split("/")
    if len(s) != 2:
        return None
    
    return s



def get_target_percentage(desired, rl_pct):
    # Get a percentage of the total amount to sell/buy
    logger.info((desired, rl_pct))
    assert 0.0 <= desired <= 1.2
    assert 0.0 <= rl_pct <= 1.0 
    return desired * rl_pct


def get_divergence_of_total(current_percentage, target_percentage):
    if target_percentage >= current_percentage:
        return 0.0
    return current_percentage - target_percentage


def get_percentage_of_sale(divergence):
    return 1-divergence


def get_sale_amount(asset_amount, percentage_to_sell):
    return asset_amount * percentage_to_sell




def buy_divergence(current_pct, target_pct):
    return target_pct-current_pct


def get_buy_pct(rl_pct, desired, current_percentage):
    # How much we're buying of the total portfolio
    target = get_target_percentage(desired=desired, rl_pct=rl_pct)
    buy_div = buy_divergence(current_pct=current_percentage, target_pct=target)
    return buy_div

def get_usd_pct(buy_pct, total_portfolio_value):
    return buy_pct*total_portfolio_value



""" The complete functions """


def get_sale_pct(decision_percentage, cap_percentage, current_percentage):
    """ 
        Dynamically get a percentage of the current amount we're trying to sell. 
    """
    # print(decision_percentage, cap_percentage, current_percentage)
    target_pct = get_target_percentage(rl_pct=decision_percentage, desired=cap_percentage)
    diverge = get_divergence_of_total(current_percentage=current_percentage, target_percentage=target_pct)
    percent_sale = get_percentage_of_sale(diverge)
    return percent_sale


def buy_total_cash(percentage, cap_allocation, pct_of_networth, net_worth):
    """ 
        # BUY TOTAL CASH
        ---
        Get the amount of cash we can spare on a given purchase in total cash. 
        Total cash = Percentage of
        ```py
            total_cash = buy_total_cash(decision_percentage, cap_allocation, pct_of_networth, net_worth)
            amount = ((total_cash / current_price)*0.98) # We get the amount of the given currency we want to buy
            trade = Trade()
        ```
        
        Parameters:
            - percentage: The percentage we want to act on
            - cap_allocation - This is the portfolio cap of the user for this given currency/portfolio. 
                - Portfolio cap is decided using rebalancing logic (modern portfolio theory)
                - float: between between 0-1
            - pct_net_worth - The percentage of the overall portfolio this asset is comprised of.
            - net_worth - The overall value of the portfolio. Renaming to networth
    """
    flexible_buy_percentage = get_buy_pct(percentage, cap_allocation, pct_of_networth)
    flexible_buy_percentage = abs(flexible_buy_percentage)
    total_cash = get_usd_pct(flexible_buy_percentage, net_worth)
    return total_cash


def get_buy_usd(decision_percentage, cap_allocation, pct_net_worth, net_worth):
    """ 
        # GET BUY USD
        ---
        Get the amount of cash we can spare on a given purchase
        buy_amount = buy_cash/price
        trade = self.buy(symbol, buy_amount, price, timestamp)
        Parameters:
            - decision_percentage: The percentage we want to act on
            - cap_allocation - This is the portfolio cap of the user for this given currency/portfolio. 
                - Portfolio cap is decided using rebalancing logic (modern portfolio theory)
                - float: between between 0-1
            - pct_net_worth - The percentage of the overall portfolio this asset is comprised of.
            - net_worth - The overall value of the portfolio. Renaming to networth
    """
    buy_pct = get_buy_pct(decision_percentage, cap_allocation, pct_net_worth)
    buy_usd = get_usd_pct(buy_pct, net_worth)
    return buy_usd



# class OrderInfo(object):
#     def get_buy_pct(self, rl_percentage, desired, current_percentage):

#         # How much we're buying of the total portfolio
#         target = get_target_percentage(desired=desired, rl_pct=rl_percentage)
#         buy_div = buy_divergence(current_pct=current_percentage, target_pct=target)
#         return buy_div
    
#     def get_sale_pct(self, rl_percentage, desired, current_percentage):
#         target_pct = get_target_percentage(rl_pct=rl_percentage, desired=desired)
#         diverge = get_divergence_of_total(current_percentage=current_percentage, target_percentage=target_pct)
#         percent_sale = get_percentage_of_sale(diverge)
        
#         # print(cy.magenta("Sales Percentage", bold=True))
#         # print("---")
#         # print(f"RL Percentage: {rl_percentage}")
#         # print(f"RL Desired: {desired}")
#         # print(f"Targeted Percentage: {target_pct}")
#         # print(f"Current percentage of total: {current_percentage}")
#         # print(f"The difference of percentage {diverge}" )
#         # print(f"The percentage of token we're holding that we intend to sell: {percent_sale}")
#         # print(rl_percentage, desired, target_pct, current_percentage, diverge, percent_sale)
#         # print()
#         return percent_sale
    
#     def get_buy_usd(self, rl_percentage, desired, current_percentage, total_portfolio_value):
#         # print(rl_percentage, desired, target_pct, current_percentage, diverge, percent_sale)
#         buy_pct = get_buy_pct(rl_percentage, desired, current_percentage)
#         buy_usd = get_usd_pct(buy_pct, total_portfolio_value)
        
#         # print(cy.magenta("Sales Percentage", bold=True))
#         # print("---")
#         # print(f"RL Percentage: {rl_percentage}")
#         # print(f"RL Desired: {desired}")
#         # print(f"Current percentage of total: {current_percentage}")
#         # print(f"Targeted Percentage: {buy_pct}")
#         # print(f"The difference of percentage {buy_usd}" )
#         return buy_usd


import copy
import random
import uuid
from typing import Optional

import maya
from jamboree import JamboreeNew as Jamboree
from jamboree.handlers.default.data import DataHandler
from loguru import logger

from see137 import PortfolioHandler, RequirementsHandler
from see137.data import PriceGenerator


class SingleAccessor(object):
    """ 
        # Simulation Creator
        ---
        Create a simulation that we'll be able to run through the faust pipeline.

        * Requirements
        * Stochastic Data
        * Portfolios
        * TimeHandler
    """
    def __init__(self, start_balance=10000, session_id:str=uuid.uuid4().hex, ecount=10, exchange="fake_exchange", **kwargs) -> None:
        self.session:str = session_id
        self.episode_count = ecount
        self.exchange = exchange
        self.start_balance = start_balance
        self.assets = ["USD_BTC", "USD_ATL", "USD_TRX", "USD_ETH", "USD_BCH", "USD_XRP", "USD_EOS", "USD_XTZ", "USD_LINK", "USD_ADA"]
        
        
        # Generated hexes
        self.episodes = [uuid.uuid4().hex for i in range(self.episode_count)]
        self.user_id = uuid.uuid4().hex # We're generating a fake user id for the time being
        # 
        self._current_episode:Optional[str] = None
        self.processor = Jamboree()
        
        # Kwargs
        self.min_price = kwargs.get("min_price", 200)
        self.max_price = kwargs.get("max_price", 5000)
        self.varied = kwargs.get("is_varied", True)
        self.length = kwargs.get("length", 5000)


        # Blanks
        self.generated = None
    
    def load_first_objects(self):
        self._portfolio = PortfolioHandler()
        self._portfolio = self.processor
        self._requirements = RequirementsHandler()
        self._requirements.processor = self.processor
        self._data = DataHandler()
        self._data.processor = self.processor

    @property
    def portfolio(self):
        """ """
        portfolio = self._portfolio
        portfolio['session'] = self.session
        portfolio['exchange'] = self.exchange
        portfolio['user_id'] = self.user_id
        portfolio['episode'] = self.current_episode
        portfolio['live'] = False
        portfolio.reset()
        return portfolio

    @property
    def requirements(self):
        requirements = self._requirements
        requirements['name'] = f"general-{self.session}"
        requirements.episode = self.current_episode
        requirements.live = False
        requirements.reset()
        return requirements
    
    @property
    def pricing(self):
        data_handler = self._data
        data_handler.processor = self.processor
        data_handler["category"] = self.category
        data_handler["subcategories"] = self.subcategories
        
        return data_handler
    
    @property
    def current_episode(self) -> str:
        if self._current_episode is None:
            return uuid.uuid4().hex
        return self._current_episode

    @current_episode.setter
    def current_episode(self, _episode:str):
        self._current_episode = _episode

    
    
    @property
    def category(self):
        return "markets"
    
    @property
    def subcategories(self):
        return {
            "market": "stock",
            "country": "US",
            "sector": "faaaaake",
            "episode": self.current_episode
        }

    @property
    def required(self):
        sub = self.subcategories
        sub.pop("episode")
        base = {
            "category": self.category,
            "subcategories": sub
        }
        ret_items = []
        for asset in self.assets:
            base = copy.copy(base)
            base['name'] = asset
            ret_items.append(base)
        return ret_items
    
    @property
    def active(self):
        """ Get all of the active requirements"""
        return self.requirements.assets

            
        


if __name__ == "__main__":
    pass
    # simulation_creator = SimulationCreator(length=2000, ecount=1)
    # simulation_creator.start_debug()

    # # Get all of the active assets in the simulation
    # active_assets = simulation_creator.active
    
    # current_active = copy.copy(active_assets)
    # random.shuffle(current_active)

    # while True:
    #     current_asset = current_active.pop()
    #     simulation_creator.requirements.report(current_asset)
    #     # with simulation_creator.requirements.lock():
    #     if simulation_creator.requirements.is_valid:
    #         logger.debug(f"All assets finished. Starting next step {current_asset}")
    #         simulation_creator.requirements.reset_reports()
            
    #         current_active = copy.copy(active_assets)
    #         random.shuffle(current_active)

    #     else:
    #         logger.error(f"Asset: {current_asset}")


import copy
import random
import uuid

import maya
from typing import Union, Optional
from addict import Dict as ADict
from jamboree import Jamboree
from jamboree.handlers.abstracted.search import ParameterizedSearch



def order_schema():
    schema = ADict()
    # exchange we're trading on
    schema.exchange = str
    # the pair we're trading on (symbol)
    schema.symbol = str
    # the amount of the base currency we're trading on
    schema.amount = float
    # the episode we're trading on. We can use to get
    schema.episode = str
    # The user the trade is started on
    schema.userid = str
    # Determines if this a live trade
    schema.live = bool
    # The side the trade is on
    schema.side = str
    # The order status (OPEN, CLOSED)
    schema.status = str
    # The time the trade initialized
    schema.timestamp = float

    # the price we initialize the trade at
    schema.init_price = float

    # The price we finished the price at
    schema.fill_price = float
    return schema.to_dict()


# PENDING = 'pending'
# OPEN = 'open'
# CANCELLED = 'cancelled'
# PARTIALLY_FILLED = 'partially_filled'
# FILLED = 'filled'


class OrderSearch(ParameterizedSearch):
    def __init__(self):
        super().__init__()
        self.entity = "order_search"
        self.dreq = order_schema()

        self.default = {"live": False, "status": "OPEN"}
        self.must_have = [
            'episode', 'userid', 'symbol', 'side', 'timestamp', 'exchange',
            'amount', 'init_price'
        ]
        self.allowed_statuses = ["PENDING", "OPEN", "CANCELLED", "PARTIAL", "FILLED"]

    def init(self, **kwargs):
        """ Starts a new order. Doesn't allow for"""
        _default = copy.copy(self.default)
        _default.update(kwargs)
        backtest_id = self.Create(no_overwrite_reqs=True, **_default)
        return backtest_id


    def generalized_search(self, _term:Union[float, str, int]="", **kwargs):
        """ A generalized search over """
        if _term is "":
            print(not kwargs)
            if not kwargs:
                return []
            return self.Find(**kwargs)
        return self.Find(general=_term, **kwargs)

    def find_by_episode(self, episode: str, **kwargs):
        """ Find of the orders inside of an episode """
        kwargs.get("episode", None)
        kwargs.update({"episode": episode})
        return self.Find(**kwargs)

    def find_open_orders(self):
        self.Find(status="OPEN")

    def find_by_exchange(self, exchange: str, **kwargs):
        kwargs.pop("exchange", None)
        kwargs.update({"exchange": exchange})
        return self.Find(**kwargs)

    def find_sim_trades(self, **kwargs):
        """ Find simulated trades """
        
        kwargs.pop("episode", None)
        kwargs.pop("live", None)
        
        simulated_schema = {"episode": '-live', "live": False}
        simulated_schema.update(kwargs)
        sim_trades = self.Find(**simulated_schema)
        print(sim_trades)
        return sim_trades

    def find_paper_trades(self, **kwargs):
        """ Finds all of the paper trades """
        
        kwargs.pop("episode", None)
        kwargs.pop("live", None)

        paper_schema = {"episode": 'live', "live": False}
        paper_schema.update(kwargs)
        print(paper_schema)
        paper_trades = self.Find(**paper_schema)
        print(paper_trades)


    def update_status(self, order_id:str, status:str, fill_price=None):
        if status not in self.allowed_statuses:
            raise ValueError(f"The status was not correct. Will allow {self.allowed_statuses}")
        if status is not "FILLED":
            # If it's not a filled status, we're just updating the order. 
            self.UpdateID(order_id, status=status)
            return

        assert fill_price != None
        self.UpdateID(order_id, status=status, fill_price=fill_price)

def main_order_scripts():
    from pprint import pprint
    _episode=uuid.uuid4().hex 
    _userid=uuid.uuid4().hex
    _exchange = random.choice(["binance", "coinbase", "okex"])
    _jam = Jamboree()
    
    order_search = OrderSearch()
    order_search.processor = _jam
    order_id = order_search.init(
        episode=_episode, 
        userid=_userid, 
        symbol="USD/BTC", 
        side="BUY",
        timestamp=maya.now()._epoch,
        exchange=_exchange,
        amount=100,
        init_price=random.uniform(0, 1000)
    )
    eps = order_search.Find(episode=_episode, exchange=_exchange)
    
    
    
    pprint(eps)
    print("\n\n")
    
    # Update the order status
    order_search.update_status(order_id, "CANCELLED")
    
    # Get the orders from this id
    # pprint(order_search.generalized_search("okex"))


if __name__ == "__main__":
    main_order_scripts()



import uuid
from typing import Optional
# from see137.exchange import price_gen_func
from see137.utils.trades import TradeType
from see137.utils.ordering import buy_total_cash, get_sale_pct

import crayons as cy

class Trade(object):
    """A trade object for use within trading environments."""

    def __init__(self, symbol: str, trade_type: 'TradeType', amount: float, price: float):
        """
        Arguments:
            symbol: The exchange symbol of the instrument in the trade (AAPL, ETH/USD, NQ1!, etc).
            trade_type: The type of trade executed (0 = HOLD, 1=LIMIT_BUY, 2=MARKET_BUY, 3=LIMIT_SELL, 4=MARKET_SELL).
            amount: The amount of the instrument in the trade (shares, satoshis, contracts, etc).
            price: The price paid per instrument in terms of the base instrument (e.g. 10000 represents $10,000.00 if the `base_instrument` is "USD").
            exchange: The exchange we're adding the trade into.
            commission: 
        """
        self._symbol = symbol
        self._trade_type = trade_type
        self._amount = amount
        self._price = price


        # Use these variables to
        self._exchange: Optional[str] = None
        self._episode: Optional[str] = None
        self._commission: Optional[str] = None
        self._userid: Optional[str] = None
        self._islive:bool = False

    def copy(self) -> 'Trade':
        """Return a copy of the current trade object."""
        return Trade(symbol=self._symbol, trade_type=self._trade_type, amount=self._amount, price=self._price)

    
    @property
    def symbol(self) -> str:
        """The exchange symbol of the instrument in the trade (AAPL, ETH/USD, NQ1!, etc)."""
        return self._symbol

    @symbol.setter
    def symbol(self, symbol: str):
        self._symbol = symbol

    @property
    def trade_type(self) -> 'TradeType':
        """The type of trade ("buy", "sell", "hold", etc)."""
        return self._trade_type

    @trade_type.setter
    def trade_type(self, trade_type: 'TradeType'):
        self._trade_type = trade_type

    @property
    def amount(self) -> float:
        """The amount of the instrument in the trade (shares, satoshis, contracts, etc)."""
        return self._amount

    @amount.setter
    def amount(self, amount: float):
        self._amount = amount

    @property
    def price(self) -> float:
        """The price paid per instrument in terms of the base instrument (e.g. 10000 represents $10,000.00 if the `base_instrument` is "USD")."""
        return float(self._price)

    @price.setter
    def price(self, price: float):
        self._price = price

    @property
    def is_hold(self) -> bool:
        """
        Returns:
            Whether the trade type is non-existent (i.e. hold).
        """
        return self._trade_type.is_hold

    @property
    def is_buy(self) -> bool:
        """
        Returns:
            Whether the trade type is a buy offer.
        """
        return self._trade_type.is_buy

    @property
    def is_sell(self) -> bool:
        """
        Returns:
            Whether the trade type is a sell offer.
        """
        return self._trade_type.is_sell
    
    @property
    def exchange(self):
        if self._exchange is None:
            raise AttributeError("Exchange hasn't been set")
        return self._exchange
    
    @exchange.setter
    def exchange(self, _exchange:str):
        self._exchange = _exchange

    @property
    def episode(self):
        if self._episode is None:
            raise AttributeError("Episode hasn't been set")
        return self._episode
    
    @episode.setter
    def episode(self, _episode:str):
        self._episode = _episode
    
    @property
    def userid(self):
        if self._userid is None:
            raise AttributeError("User id not set")
        return self._userid
    
    @userid.setter
    def userid(self, user_id):
        self._userid = user_id

    @property    
    def live(self) -> bool:
        return self._islive

    @live.setter
    def live(self, _live:bool):
        self._live = _live

    def __str__(self) -> str:
        action_name = "BUY"
        if self.is_sell:
            action_name = "SELL"
        elif self.is_hold:
            action_name = "HOLD"
        
        return_string = f"<Trade: {action_name}, symbol:{self.symbol}, amount:{self.amount}, price: {self.price}>"
        return return_string

class DynamicTrade(object):
    """A trade object for use within trading environments."""

    def __init__(self, symbol: str, trade_type: 'TradeType', percentage=0.4, cap=0.5, pct_net_worth=0.3, net_worth=10000):
        """
        Arguments:
            symbol: The exchange symbol of the instrument in the trade (AAPL, ETH/USD, NQ1!, etc).
            trade_type: The type of trade executed (0 = HOLD, 1=LIMIT_BUY, 2=MARKET_BUY, 3=LIMIT_SELL, 4=MARKET_SELL).
            amount: The amount of the instrument in the trade (shares, satoshis, contracts, etc).
            price: The price paid per instrument in terms of the base instrument (e.g. 10000 represents $10,000.00 if the `base_instrument` is "USD").
        """
        self._symbol = symbol
        self._trade_type = trade_type
        self._percentage = percentage
        self._net_worth = net_worth
        self._cap_pct = cap
        self._current_pct_net_worth = pct_net_worth
        self._amount = 0.0
        self._price = 0.0000001

    def copy(self) -> 'Trade':
        """Return a copy of the current trade object."""
        return Trade(symbol=self._symbol, trade_type=self._trade_type, price=self._price, amount=self._amount)
    
    def calculate_trade(self) -> 'Trade':
        """Return a copy of the current trade object."""
        absolute_amount = self.amount * self.price
        self.current_pct = (absolute_amount/self.net_worth)
        if self.is_buy:
            total_cash = buy_total_cash(
                percentage=self.percentage, 
                cap_allocation=self.cap_pct, 
                pct_of_networth=self.current_pct, 
                net_worth=self.net_worth
            )
            if self.price != 0.0: 
                self.amount = ((total_cash/self.price) * 0.99999998)
            else:
                self.amount = 0.0
            
        elif self.is_sell:
            percentage_to_sell = get_sale_pct(
                decision_percentage=self.percentage,
                cap_percentage=self.cap_pct,
                current_percentage=self.current_pct
            )
            # Amount should already be set by now to the number of items
            trade_amount = ((self.amount * percentage_to_sell) * 0.99999998)
            self.amount = trade_amount

        return Trade(symbol=self.symbol, trade_type=self.trade_type, price=self.price, amount=self.amount)

    
    @property
    def symbol(self) -> str:
        """The exchange symbol of the instrument in the trade (AAPL, ETH/USD, NQ1!, etc)."""
        return self._symbol

    @symbol.setter
    def symbol(self, symbol: str):
        self._symbol = symbol

    @property
    def trade_type(self) -> 'TradeType':
        """The type of trade ("buy", "sell", "hold", etc)."""
        return self._trade_type

    @trade_type.setter
    def trade_type(self, trade_type: 'TradeType'):
        self._trade_type = trade_type

    @property
    def amount(self) -> float:
        """The amount of the instrument in the trade (shares, satoshis, contracts, etc)."""
        return self._amount

    @amount.setter
    def amount(self, amount: float):
        self._amount = amount

    @property
    def price(self):
        return self._price
    
    @price.setter
    def price(self, _price:float):
        self._price = _price

    @property
    def is_hold(self) -> bool:
        """
        Returns:
            Whether the trade type is non-existent (i.e. hold).
        """
        return self._trade_type.is_hold

    @property
    def is_buy(self) -> bool:
        """
        Returns:
            Whether the trade type is a buy offer.
        """
        return self._trade_type.is_buy

    @property
    def is_sell(self) -> bool:
        """
        Returns:
            Whether the trade type is a sell offer.
        """
        return self._trade_type.is_sell


    """ Buy and sell properties """

    @property
    def percentage(self):
        return self._percentage
    

    @percentage.setter
    def percentage(self, _pct):
        self._percentage = _pct

    @property
    def cap_pct(self):
        return self._cap_pct


    @cap_pct.setter
    def cap_pct(self, _cap_pct):
        self._cap_pct = _cap_pct
    
    @property
    def current_pct(self) -> float:
        """ The currently held percentage of the total value """
        return self._current_pct_net_worth
    

    @current_pct.setter
    def current_pct(self, _pct):
        """ The currently held percentage of the total value """
        self._current_pct_net_worth = _pct
        
    
    @property
    def net_worth(self):
        return self._net_worth
    
    @net_worth.setter
    def net_worth(self, _worth):
        self._net_worth = _worth


    def __str__(self) -> str:
        current_trade = self.calculate_trade()
        return str(current_trade)


if __name__ == "__main__":
    dynamic_trade = DynamicTrade("BTC", TradeType.LIMIT_BUY)
    dynamic_trade.price = 200
    dynamic_trade.cap_pct = 0.35
    dynamic_trade.percentage = 1.0
    trade = dynamic_trade.calculate_trade()
    print(trade)




import uuid
import maya
import random
from cytoolz.curried import pipe, map, filter, get
from cytoolz.itertoolz import unique

from typing import Optional, List
from jamboree import Jamboree
from jamboree.base.processors.abstracts import Processor
from jamboree.handlers.abstracted.search import ParameterizedSearch
from jamboree.handlers.default.search import BaseSearchHandler
from see137.search.prototype import SearchFacade
from loguru import logger

class _BacktestUniverseManagement(BaseSearchHandler):
    """ 
        # Changing this method to store all relavent universe information.

        Things like descriptions, userid, name, extra tags. Use to optimize query to a single component when reading.
        
        
        ```
    """
    def __init__(self):
        super().__init__()
        self.entity = "universe_information"
        # actually, let's say the multi-id and universe id are separate 
        self.dreq = {"universe": str, "name": str, "description": str, "user":str, "tags":list}

class BacktestUniverseManagement(SearchFacade):
    def __init__(self):
        super().__init__(_BacktestUniverseManagement)


class UniverseCreate(BacktestUniverseManagement):
    """ 
        Used to find stategies and universes by asset or universe. 

        Store all relavent information regarding the user/universe interaction. Use to denormalize

        ```py
            # Match a strategy with a universe and asset
            backtest_manager.create.universe.init(universe: str, strategy: str, asset:str)
        ```
    """
    def __init__(self):
        super().__init__()
        self.must_have = ["universe"]
    
    def init(self, universe: str, name:str, user:str, description:str, tags:List[str]=[]):
        if bool(tags):
            universe_mix_id = self.Create(no_overwrite_reqs=True, universe=universe, name=name, user=user, description=description, tags=tags)
        return universe_mix_id

    
    
class UniverseUpdate(BacktestUniverseManagement):
    """ 
        Use to find backtest information by episodes, and episodes by backtest

        ```py
            # Start a new backtest
            backtest_manager.create.universe.init(backtest_id)

            
        ```
    """
    def __init__(self):
        super().__init__()
        
    
    def name(self, universe:str, _name:str):
        """ Update episode status """
        self.UpdateID(universe, name=_name)
    
    def description(self, universe:str, description:str):
        self.UpdateID(universe, description=description)
    

        

    


class UniverseInfo(BacktestUniverseManagement):
    """ 
        Use to find backtest information by episodes, and episodes by backtest

        - Is there any backtest that's running right now?
        - Find all of the unique universes

        ```py
            # Get the difference
            backtest_manager.info.episodes.by_backtest(backtest_id)
            
        ```
    """
    def __init__(self):
        super().__init__()
    
    @property
    def universes(self) -> List[dict]:
        """ get all episodes """
        universes = self.Find("*")
        unis = pipe(universes, map(lambda uni: uni['universe']), unique, list)
        return unis
    
    @property
    def assets(self) -> List[dict]:
        """ Get all active episodes"""
        assets = self.Find("*")
        unis = pipe(assets, map(lambda uni: uni['asset']), unique, list)
        return unis
    
    @property
    def strategies(self) -> List[str]:
        """ Return a list of list of strategy IDs"""
        strategies = self.Find("*")
        unis = pipe(strategies, map(lambda uni: uni['strategy']), unique, list)
        return unis



    @property
    def total_universes(self) -> int:
        return len(self.universes)

    @property
    def total_strategies(self) -> int:
        return len(self.strategies)
    
    @property
    def total_assets(self) -> int:
        return len(self.assets)

    
    def universes_by_user(self, user:str):
        """ Get all of the created universes by user. """
        strategy_list = self.Find(user=user)
        strats = pipe(strategy_list, map(lambda uni: uni['universe']), unique, list)
        return strats
    
    def by_user(self, user:str):
        """ Get all records by user."""
        strategy_list = self.Find(user=user)
        return strategy_list

    
    def asset_by_universe(self, universe:str):
        assets = self.Find(universe=universe)
        _assets = pipe(assets, map(lambda assss: assss['asset']), unique, list)
        return _assets
    
    def by_fuzzy_description(self, description:str):
        if bool(description):
            records = self.Find(description=f"{description}*")
            return records
        return []

class UniverseRemove(BacktestUniverseManagement):
    """ 
        
    """
    def __init__(self):
        super().__init__()

    def by_strat_uni(self, strategy:str, universe:str):
        """ Remove episode by id"""
        self.Remove(strategy=strategy, universe=universe)

    def by_strat_asset(self, strategy:str, asset:str):
        self.Remove(strategy=strategy, asset=asset)


def manage_assets():
    strat_id = uuid.uuid4().hex
    universe_ids = [uuid.uuid4().hex for x in range(10)]
    asset_ids = [uuid.uuid4().hex for x in range(200)]
    
    proc = Jamboree()
    universe_create =    UniverseCreate()
    universe_update =    UniverseUpdate()
    universe_info =      UniverseInfo()
    universe_remove =    UniverseRemove()
    
    universe_create.processor = proc
    universe_update.processor = proc
    universe_info.processor = proc
    universe_remove.processor = proc

    # for assetid in asset_ids:
    #     for _ in range(10):
    #         universeid = random.choice(universe_ids)
    #         strat_uni_id = universe_create.init(universeid, strat_id, assetid)
    #     strategies = universe_info.strategy_by_asset(assetid)
    #     for x in strategies:
    #         logger.success(f"Sending command to strategy: {x}. Looking for strategy.") 
    #         logger.debug(assetid)

    #     logger.error(strat_uni_id)
    #     assets = universe_info.asset_by_universe(universeid)
    #     logger.warning(assets)

if __name__ == "__main__":
    manage_assets()


from see137.parameters import Parameter
from jamboree.utils.core import consistent_hash

class UserParameter(Parameter):
    """ UserConfig """
    def __init__(self, userid:str, is_websocket=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user = userid
        self.is_websocket = is_websocket
        
        # We're eventually allowing to point to another portfolio to begin the backtest on (start amount, allocation, etc) 
    


import datetime as dt
import random
import time
import uuid
from typing import Dict, List

from addict import Dict as ADict
from jamboree import Jamboree
from jamboree.handlers.default.multi import MultiDataManagement
from jamboree.handlers.default.search import BaseSearchHandler
from loguru import logger

from see137 import PortfolioHandler, RequirementsHandler
from see137.common import (AddictPipeline, AddictWithManagerComponent,
                                    ComponentAbstract, Pipeline, PipelineA)
from see137.components.backtest import CreateBacktestComponent
from see137.components.strategy import CheckStrategy
from see137.components.universe import CheckUniverse
from see137.handlers.strategy.core import StrategyEngine
from see137.models.backtests_models import BacktestConfig
from see137.search.prototype import SearchFacade
from see137.search.backtesting.topics.specifics import BacktestEpisodeSpec


class CreateRequirements(AddictWithManagerComponent):
    def __init__(self):
        super().__init__()
        # Backtest configuration dict. Can be saved into the database immediately.
        self.required_items: Dict[str, type] = {"config": dict}
        self.__private_multi = MultiDataManagement()
        self.__private_requirements = RequirementsHandler()
        self.mult_handler = None

    @property
    def requirements(self):
        self.__private_requirements.processor = self.processor
        self.__private_requirements.live = False
        return self.__private_requirements

    def check_prev(self, item):
        return super().check_prev(item)

    def check_all(self, item):
        return super().check_all(item)

    def check_item(self, item):
        return super().check_item(item)

    @property
    def multi(self):
        self.__private_multi.processor = self.processor
        return self.__private_multi

    def process(self, item: ADict):
        """Steps:

            1. Pull universe
            2. Verify it

        Arguments:
            item {ADict} -- Addict modified dictionary. Save's according to backtest data.

        Returns:
            ADict -- Modified dictionary with action and percentage inside of it
        """
        universe = item.config.universe
        logger.debug(f"Trying to pull the universe: {universe}")
        universe_handler = self.multi.pick(universe)
        if universe_handler is None:
            raise AttributeError("Universe doesn't exist")

        asset_list = universe_handler.sources
        asset_suc_msg = f"Successfully pulled universe of assets for {universe}"
        logger.success(asset_suc_msg)
        general_req_name = f"general_universe_{universe}"
        item.requirementid = general_req_name
        self.requirements['name'] = general_req_name
        self.requirements.live = False
        self.requirements.reset()
        self.requirements.asset_update(asset_list)
        self.requirements.episode = item.current_episode

        episodic_requirements = self.requirements.unique()
        msg = f"UNIQUE REQUIREMENTS {episodic_requirements}"
        logger.warning(msg)
        return item


class AttachToEpisode(AddictWithManagerComponent):
    """Attaches everything from the prior steps to the episode:

    1. Requirement name
    2. StrategyID
    3. Aggregate Portfolio ID

    Arguments:
        AddictWithManagerComponent {[type]} -- Basically a component that uses addict as a foundation
    """
    def __init__(self):
        super().__init__()
        # Backtest configuration dict. Can be saved into the database immediately.
        self.required_items: Dict[str, type] = {
            "config": dict,
            "requirementid": str,
            "current_episode": str
        }
        self.__private_epispec = BacktestEpisodeSpec()
        self.__private_multi = MultiDataManagement()
        self.__private_requirements = RequirementsHandler()
        self.mult_handler = None

    @property
    def requirements(self):
        self.__private_requirements.processor = self.processor
        self.__private_requirements.live = False
        return self.__private_requirements

    @property
    def spec(self):
        self.__private_epispec.processor = self.processor
        return self.__private_epispec

    def check_prev(self, item):
        return super().check_prev(item)

    def check_all(self, item):
        return super().check_all(item)

    def check_item(self, item):
        return super().check_item(item)

    @property
    def multi(self):
        self.__private_multi.processor = self.processor
        return self.__private_multi

    def process(self, item: ADict):
        """Steps:

            1. Pull universe
            2. Verify it

        Arguments:
            item {ADict} -- Addict modified dictionary. Save's according to backtest data.

        Returns:
            ADict -- Modified dictionary with action and percentage inside of it
        """
        universe = item.config.universe
        strategy = item.config.strategy
        backtest = item.backtestid
        episode = item.current_episode

        save_ref = self.spec.Create(no_overwrite_reqs=True,
                                    universe=universe,
                                    strategy=strategy,
                                    backtest=backtest,
                                    episode=episode)
        logger.error(f"The save reference is: {save_ref}")
        return item




import math
import numpy
import numpy.random as nrand


def convert_to_returns(log_returns):
    """
    This method exponentiates a sequence of log returns to get daily returns.
    :param log_returns: the log returns to exponentiated
    :return: the exponentiated returns
    """
    return numpy.exp(log_returns)


def convert_to_prices(param, log_returns):
    """
    This method converts a sequence of log returns into normal returns (exponentiation) and then computes a price
    sequence given a starting price, param.all_s0.
    :param param: the model parameters object
    :param log_returns: the log returns to exponentiated
    :return:
    """
    returns = convert_to_returns(log_returns)
    # A sequence of prices starting with param.all_s0
    price_sequence = [param.all_s0]
    for i in range(1, len(returns)):
        # Add the price at t-1 * return at t
        price_sequence.append(price_sequence[i - 1] * returns[i - 1])
    return numpy.array(price_sequence)

def brownian_motion_log_returns(param):
    """
    This method returns a Wiener process. The Wiener process is also called Brownian motion. For more information
    about the Wiener process check out the Wikipedia page: http://en.wikipedia.org/wiki/Wiener_process
    :param param: the model parameters object
    :return: brownian motion log returns
    """
    sqrt_delta_sigma = math.sqrt(param.all_delta) * param.all_sigma
    return nrand.normal(loc=0, scale=sqrt_delta_sigma, size=param.all_time)


import uuid
import inspect
from loguru import logger
from typing import List, Dict, Tuple, Any
from addict import Dict as ADict
from see137.common import ComponentAbstract


class AddictComponent(ComponentAbstract):
    """
        A component that centers around using dictionaries and addict.
    """
    def __init__(self):
        self.name = self.__class__.__name__
        
        self.within_pipeline = None
        self.required_items:Dict[str, type] = {}
        self.required_settings: Dict[str, type] = {}
        self.settings:ADict = ADict()
        self.is_implemented = False
        
    

    def set_settings(self, **kwargs):
        # self.settings.is_loud = kwargs.get("is_loud", False)
        failed = False
        msg = f"All settings here there for {self.name}"
        if not bool(self.required_settings):
            return failed, "No requirement items to check. Processing ..."

        keys = list(kwargs.keys())
        req_keys = (self.required_settings.keys())
        if len(keys) < len(req_keys):
            self.raise_loud(AttributeError, f"You need ALL of the following keys: {req_keys}. You have: {keys}")
            return True, msg
        try:
            for key, value in self.required_settings.items():
                cap = key.capitalize()
                # ck = key
                current = kwargs[key]
                if key not in keys:
                    raise ValueError(f"{cap} not in the kwargs. ")
                if not isinstance(current, value):
                    raise TypeError(f"{cap} not correct instance type {str(value)}. ")
                self.settings[key] = current
        except KeyError as ke:
            self.raise_loud(KeyError, f"You haven't added the key {str(ke)} to settings")
            return True, str(e)
        except Exception as e:
            self.raise_loud(e, str(e))
            return True, str(e)
        return failed, msg
    

    def check_prev(self, item:ADict):
        """Checks to see if the previous component failed. If it did we skip this task.

        Arguments:
            item {ADict} -- The previous item
        """

        if not bool(item.failed):
            item.failed = False
            item.msg = "Previous item succeeded. Processing here ..."


        if item.failed:
            item.failed = True
            if not bool(item.msg):
                item.msg = f"Previous Component Failed"
            self.raise_loud(ValueError, item.msg)
            
        return item
        
        

    def check_item(self, item:ADict) -> Tuple[bool, str]:
        failed = False
        msg = "All required items pass"
        if not bool(self.required_items):
            return failed, "No items to check."
        
        for key, value in self.required_items.items():
            current = item[key]
            is_valid = bool(current)
            is_bool = isinstance(current, bool)
            if not is_valid and not is_bool:
                failed = True
                msg = f"{key} field not found in the dictionary"
                self.raise_loud(TypeError, msg)
                break
            if value == Any:
                self.settings[key] = value
                continue
            if not isinstance(item[key], value):
                failed = True
                msg = f"{key} field not the correct type"
                self.raise_loud(TypeError, msg)
                break
            self.settings[key] = value
        return failed, msg
    
    def check_all(self, item:ADict):
        item = self.check_prev(item)
        
        if item.failed:
            return item
        failed, msg = self.check_item(item)
        if failed:
            item.failed = failed
            item.msg = msg

        return item
    
    def process(self, item:ADict):
        """Process component using procedure defined here. 

        Arguments:
            item {ADict} -- A modified dictionary with the results inputted in.
        """
        return item


    def raise_loud(self, exc=Exception, msg="General Problem"):
        if self.settings.loud:
            msg = f"{msg}. Error at component {self.name} ... Pipeline: {self.within_pipeline}"
            raise exc(msg)

    def step(self, item:ADict, **kwargs):
        self.settings.loud=kwargs.pop("is_loud", True)
        self.settings.override=kwargs.pop("override", False)
        failed, msg = self.set_settings(**kwargs)
        

        if failed:
            item.failed = failed
            item.msg = msg
            self.raise_loud(Exception, item.msg)

        # self.settings.loud = kwargs.get("is_loud", False)
        item = self.check_all(item)
        if item.failed:
            self.raise_loud(Exception, item.msg)
            return item
        processed = self.process(item)
        return processed

if __name__ == "__main__":
    comp = AddictComponent()
    sample_param = ADict(episode=uuid.uuid4().hex)
    print(comp.step(sample_param, asset=uuid.uuid4().hex))


# You're creating an ORM. Look at what's already there
# BaseModel
from typing import List, Union, Dict, Any
from pydantic import BaseModel

class Foo(BaseModel):
    data: Dict[str, Union[str, bool, int, float, Dict[str, Any]]] = {}
    size: float = None

    def field_keys(self) -> List[str]:
        return list(self.fields.keys())

    def create_query(self):
        print(self.field_keys())
        print("Hello World")


Foo().create_query()



from typing import Optional
from loguru import logger
from jamboree import DBHandler


class MetaProperty:
    """
        NOTE: Needs to have schemas to bring out the full power of this.
        This is to interact with the properties of the individual items with a single builder pattern. 
        It should reduce amount of code directly on each object.
    """


    def __init__(self) -> None:
        self._alt = {}
        self.dbhand:Optional[DBHandler] = None
        self._schema = {}
        self._default_type_mapping = {
            int: 0,
            float: 0.0,
            str: "",
            dict: {},
            list: [],
        }


    @property
    def schema(self) -> dict:
        return self._schema

    @schema.setter
    def schema(self, _schema:dict):
        if self.is_all_type_vals(_schema):
            self._schema = _schema

    @property
    def is_schema(self) -> bool:
        lll = len(self.schema.keys())
        if lll == 0: False 
        return True


    @property
    def default_dict(self):
        schema_keys = list(self.schema.keys())
        default_keys = list(self._default_type_mapping.keys())
        defaults = {}
        if len(schema_keys) == 0: return defaults
        for skey, sval in self.schema.items():
            if sval not in default_keys:
                defaults[skey] = 0
            else:
                defaults[skey] = self._default_type_mapping[sval]
        return defaults


    def is_all_type_vals(self, _schema:dict):
        for sval in _schema.values():
            type_cast = type(sval)
            is_typing = isinstance(type_cast, type)
            if is_typing:
                return False
        return True

    def validate_schema(self, item:dict) -> bool:
        if self.is_schema: return True
        for skey, sval in self.schema.items():
            if skey not in item:
                return False
            if not isinstance(item[skey], sval):
                return False
        return True
    
    

    def check_for_methods(self):
        if self.dbhand is None:
            raise AttributeError("Didn't inject the database handler")

        if not hasattr(self.dbhand, "save") or not hasattr(self.dbhand, "processor"):
            raise AttributeError("The database handler doesn't have the appropiate functions")
    
    
    def prop(self, _alt):
        self.check_for_methods()
        self._alt = _alt
        return self

    def last(self) -> dict:
        """ Get the last of a given item"""
        return self.dbhand.last(self._alt)

    def count(self) -> int:
        return self.dbhand.count(self._alt)
    
    def reset(self):
        if self.dbhand.count(self._alt) == 0:
            default = self.default_dict
            self.dbhand.save(default, alt=self._alt)





import maya
import uuid
import dask
import copy
import random
import pandas as pd
import numpy as np
from loguru import logger
from crayons import (
    red, magenta, yellow, cyan, green
)
from typing import Optional, List

from see137.stochastic import generate_super_price
from jamboree import JamboreeNew

from jamboree.handlers.default.data import DataHandler
from jamboree.handlers.default.multi import MultiDataManagement

class PriceGenerator(object):
    def __init__(self, **kwargs) -> None:
        self._start_time = maya.now()
        self._seconds = kwargs.get("seconds", 0)
        self._minutes = kwargs.get("minutes", 0)
        self._hours = kwargs.get("hours", 0)
        self._days = kwargs.get("days", 1)
        self._months = kwargs.get("months", 0)
        self._years = kwargs.get("years", 0)
        
        self.asset_bars = {}
        self._assets = set()
        self._episodes = set()
        
        self.starting_min=50
        self.starting_max=2000
        self.is_varied=True
        self.excluded_assets=[]
        self.length=4000
        self.episode_set = {}


    @property
    def assets(self):
        return list(self._assets)
    
    @assets.setter
    def assets(self, _assets:list):
        if len(_assets) == 0:
            return
        for _ in _assets:
            self._assets.add(_)


    @property
    def start_time(self) -> 'maya.MayaDT':
        return copy.copy(self._start_time)
    
    
    
    @start_time.setter
    def start_time(self, _start_time):
        if isinstance(_start_time, float):
            # we know it's an epoch
            self._start_time = maya.MayaDT(_start_time)
        
        elif isinstance(_start_time, maya.MayaDT):
            self._start_time = _start_time
    

    @property
    def episodes(self):
        """ """
        return list(self._episodes)
    
    @episodes.setter
    def episodes(self, _episodes:List[str]):
        """ Add a list of episodes """
        if len(_episodes) == 0:
            return
        for episode in _episodes:
            self._episodes.add(episode)

    
    def generate_price_bar(self, name:str, base_type:Optional[str]=None, base_len:int=1000, base_price:float=1000.0):
        base_type = base_type or random.choice(
            ["GBM", "HESTON", "MERTON", "COX", "ORNSTEIN"]
        )
        price = generate_super_price(
            base_price=base_price,
            base_type=base_type,
            base_time=base_len
        )
        ps = len(price)
        vol_multiplier = random.uniform(2, 20)
        open_multiplier = np.absolute(np.random.normal(1, 0.03, size=ps))
        high_multiplier = np.absolute(np.random.normal(1.12, 0.03, size=ps))
        low_multiplier =  np.absolute(np.random.normal(0.92, 0.03, size=ps))
        noise = np.absolute(
            np.random.normal(1, 0.2, size=ps)
        )
    
        volume_bars = (noise * price) * vol_multiplier
        open_prices = price * open_multiplier
        high_prices = high_multiplier * price
        low_prices = low_multiplier * price

        bars = []
        start_time = self.start_time
        count = 0
        for _ in range(ps):
            bar = {
                "open": float(open_prices[_]),
                "close": float(price[_]),
                "high": float(high_prices[_]),
                "volume": float(volume_bars[_]),
                "low": float(low_prices[_]),
                "time": float((start_time.add(
                                seconds=(self._seconds * count), 
                                minutes=(self._minutes * count), 
                                hours=(self._hours * count), 
                                days=(self._days * count), 
                                months=(self._months * count), 
                                years=(self._years * count)
                )._epoch))
            }
            count += 1
            bars.append(bar)
        
        frame = pd.DataFrame(bars)
        timestamps = pd.to_datetime(frame.time, unit='s')
        frame.set_index(timestamps, inplace=True)
        frame = frame.drop(columns=["time"])
        self.asset_bars[name] = frame


    def generate(self, starting_min=50, starting_max=2000, is_varied=True, excluded_assets=[], _len=4000):
        """ Generate price bars for assets """
        assets = self.assets
        if len(assets) == 0:
            logger.error(red("No assets found"))
            return
        
        for excluded in self.excluded_assets:
            assets.remove(excluded)
        
        # Generate bars

        dask_tasks = []
        for asset in assets:
            base_price = random.uniform(self.starting_min, self.starting_max)
            _type = random.choice(
                ["GBM", "HESTON", "MERTON", "ORNSTEIN"]
            )
            dask_task = dask.delayed(self.generate_price_bar)(asset, _type, self.length, base_price)
            dask_tasks.append(dask_task)
        dask.compute(*dask_tasks)
        logger.success(f"All pricing successfully generated for episode")
        return self.asset_bars

    def generate_episode_and_set(self, episode):
        pricing_generation = self.generate()
        self.episode_set[episode] = pricing_generation
        logger.success("Finished generated episode")

    def generate_all_episodes(self):
        self.episode_set = {}
        dask_tasks = []
        for episode in self.episodes:
            dask_task = dask.delayed(self.generate_episode_and_set)(episode)
            # pricing_generation = self.generate()
            # self.episode_set[episode] = pricing_generation
            dask_tasks.append(dask_task)
        dask.compute(*dask_tasks)
        return self.episode_set

def create_data_item(category:str, subcategory:dict, name:str):
    dset = {
        "name": name,
        "subcategories": subcategory,
        "category": category
    }
    return dset

def price_gen_func():
    asset_list = ["BTC", "ATL", "TRX", "ETH", "BCH", "XRP"]
    episodes = [uuid.uuid4().hex for x in range(10)]
    price_gen = PriceGenerator()
    price_gen.episodes = episodes
    price_gen.assets = asset_list
    price_gen.starting_min = 50
    price_gen.starting_max=2000
    price_gen.is_varied=True
    price_gen.length=2000
    price_gen.generate()
    return price_gen

def price_gen_funcs(
        number_of_eps:int, 
        asset_list:List[str]=["BTC", "ATL", "TRX", "ETH", "BCH", "XRP"]) -> 'PriceGenerator':
    episodes = [uuid.uuid4().hex for x in range(number_of_eps)]
    price_gen = PriceGenerator()
    price_gen.episodes = episodes
    price_gen.assets = asset_list
    price_gen.starting_min = 50
    price_gen.starting_max=2000
    price_gen.is_varied=True
    price_gen.length=1000
    price_gen.generate_all_episodes()
    return price_gen

def set_heads(episodes:List[str], data_handler:'DataHandler', start_time:maya.MayaDT):
    if len(episodes) <= 0:
        return 
    for episode in episodes:
        logger.info(cyan(episode))
        data_handler.episode = episode
        data_handler.time.head = start_time.add(weeks=4)._epoch
        data_handler.time.change_stepsize(microseconds=0, days=1, hours=0)
        data_handler.time.change_lookback(microseconds=0, weeks=4, hours=0)



def forward_run(episodes:List[str], data_handler:'DataHandler', assets:List[str]):
    copied_episodes = copy.copy(episodes)
    copied_assets = copy.copy(assets)
    while True:
        for episode in copied_episodes:
            data_handler.episode = episode
            for asset in copied_assets:
                data_handler['name'] = asset
                if data_handler.is_next:
                    data_handler.time.step()
                else:
                    continue

def main():
    jam_processor = JamboreeNew()
    data_handler = DataHandler()
    multi_data = MultiDataManagement()
    data_handler.processor = jam_processor
    multi_data.processor = jam_processor
    episode = uuid.uuid4().hex

    start = maya.now()._epoch
    price_gen = price_gen_funcs(10)

    
    data_handler.live = False
    category = "fake_markets"
    subcategories = {
        "market": "stock",
        "country": "US",
        "sector": "faaaaake"
    }

    multi_data["set_name"] = uuid.uuid4().hex
    data_handler['category'] = category
    data_handler['subcategories'] = subcategories



    for episode in price_gen.episodes:
        all_coins = price_gen.episode_set[episode]
        coin_keys = list(all_coins['data'].keys())

        if len(coin_keys) == 0:
            return
        
        data_handler.episode = episode
        for coin in coin_keys:
            data = all_coins['data'][coin]
            data_handler['name'] = coin
            data_handler.reset()
            data_handler.store_time_df(data)
    

    set_heads(price_gen.episodes, data_handler, price_gen.start_time)
    forward_run(price_gen.episodes, data_handler, price_gen.assets)



if __name__ == "__main__":
    main()


import math
import numpy
import random
import decimal
import scipy.linalg
import numpy.random as nrand
import matplotlib.pyplot as plt
import pandas as pd
from datetime import datetime


class ModelParameters:
    """
    Encapsulates model parameters
    """

    def __init__(self,
                 all_s0, all_time, all_delta, all_sigma, gbm_mu,
                 jumps_lamda=0.0, jumps_sigma=0.0, jumps_mu=0.0,
                 cir_a=0.0, cir_mu=0.0, all_r0=0.0, cir_rho=0.0,
                 ou_a=0.0, ou_mu=0.0,
                 heston_a=0.0, heston_mu=0.0, heston_vol0=0.0):
        # This is the starting asset value
        self.all_s0 = all_s0
        # This is the amount of time to simulate for
        self.all_time = all_time
        # This is the delta, the rate of time e.g. 1/252 = daily, 1/12 = monthly
        self.all_delta = all_delta
        # This is the volatility of the stochastic processes
        self.all_sigma = all_sigma
        # This is the annual drift factor for geometric brownian motion
        self.gbm_mu = gbm_mu
        # This is the probability of a jump happening at each point in time
        self.lamda = jumps_lamda
        # This is the volatility of the jump size
        self.jumps_sigma = jumps_sigma
        # This is the average jump size
        self.jumps_mu = jumps_mu
        # This is the rate of mean reversion for Cox Ingersoll Ross
        self.cir_a = cir_a
        # This is the long run average interest rate for Cox Ingersoll Ross
        self.cir_mu = cir_mu
        # This is the starting interest rate value
        self.all_r0 = all_r0
        # This is the correlation between the wiener processes of the Heston model
        self.cir_rho = cir_rho
        # This is the rate of mean reversion for Ornstein Uhlenbeck
        self.ou_a = ou_a
        # This is the long run average interest rate for Ornstein Uhlenbeck
        self.ou_mu = ou_mu
        # This is the rate of mean reversion for volatility in the Heston model
        self.heston_a = heston_a
        # This is the long run average volatility for the Heston model
        self.heston_mu = heston_mu
        # This is the starting volatility value for the Heston model
        self.heston_vol0 = heston_vol0
    
def convert_to_returns(log_returns):
    """
    This method exponentiates a sequence of log returns to get daily returns.
    :param log_returns: the log returns to exponentiated
    :return: the exponentiated returns
    """
    return numpy.exp(log_returns)


def convert_to_prices(param, log_returns):
    """
    This method converts a sequence of log returns into normal returns (exponentiation) and then computes a price
    sequence given a starting price, param.all_s0.
    :param param: the model parameters object
    :param log_returns: the log returns to exponentiated
    :return:
    """
    returns = convert_to_returns(log_returns)
    # A sequence of prices starting with param.all_s0
    price_sequence = [param.all_s0]
    for i in range(1, len(returns)):
        # Add the price at t-1 * return at t
        price_sequence.append(price_sequence[i - 1] * returns[i - 1])
    return numpy.array(price_sequence)



def brownian_motion_log_returns(param):
    """
    This method returns a Wiener process. The Wiener process is also called Brownian motion. For more information
    about the Wiener process check out the Wikipedia page: http://en.wikipedia.org/wiki/Wiener_process
    :param param: the model parameters object
    :return: brownian motion log returns
    """
    sqrt_delta_sigma = math.sqrt(param.all_delta) * param.all_sigma
    return nrand.normal(loc=0, scale=sqrt_delta_sigma, size=param.all_time)


def brownian_motion_levels(param):
    """
    Returns a price sequence whose returns evolve according to a brownian motion
    :param param: model parameters object
    :return: returns a price sequence which follows a brownian motion
    """
    return convert_to_prices(param, brownian_motion_log_returns(param))





def is_acceptable_type(_type:str) -> bool:
    """ Determine if we're putting an appropiate stochastic generated type """
    upper = _type.upper()

    if upper not in ["GBM", "HESTON", "MERTON", "COX", "ORNSTEIN"]:
        return False
    
    return True

def get_super_parameters(base_price, base_time):
    if base_price < 0.05:
        raise ValueError("The base price needs to be higher than 0.05 to make everything work.")

    if base_time < 50:
        raise ValueError("We need to simulate at least 50 points")

    delta = (1/252)
    super_params = ModelParameters(
        all_s0=base_price,
        all_r0=0.5,
        all_time=base_time,
        all_delta=delta,
        all_sigma=random.uniform(0.01, 0.4),
        gbm_mu=random.uniform(0.05, 0.1),
        jumps_lamda=random.uniform(0.01, 0.03),
        jumps_sigma=random.uniform(0.01, 0.2),
        jumps_mu=random.uniform(0.0001, 0.007),
        cir_a=3.0,
        cir_mu=0.5,
        cir_rho=0.5,
        ou_a=random.normalvariate(3.0, 0.4),
        ou_mu=random.normalvariate(0.5, 0.2),
        heston_a=random.uniform(3, 5),
        heston_mu=random.uniform(0.156, 0.693),
        heston_vol0=0.06125
    )
    return super_params


import uuid
import maya

from typing import Optional, List
from jamboree import Jamboree
from jamboree.base.processors.abstracts import Processor
from jamboree.handlers.abstracted.search import ParameterizedSearch
from jamboree.handlers.default.search import BaseSearchHandler
from see137.search.prototype import SearchFacade
from loguru import logger

class AggregatePortfolio(BaseSearchHandler):
    """A search handler the aggregate portfolio

    Arguments:
        BaseSearchHandler {[type]} -- [description]
    """
    def __init__(self):
        super().__init__()

        self.entity = "aggregate_portfolio"
        self.dreq = {
            'timestamp': float,
            'episode': str,
            "user": str,
            'total': float,
            "allocation": dict,
            'performance': float
        }

class BacktestPortfolioAgg(SearchFacade):
    def __init__(self):
        ProcedureSearch = AggregatePortfolio()
        super().__init__(ProcedureSearch)
        self.must_have = ['episode']


from enum import Enum


class TradeType(Enum):
    """A trade type for use within trading environments."""

    HOLD = 0
    LIMIT_BUY = 1
    MARKET_BUY = 2
    LIMIT_SELL = 3
    MARKET_SELL = 4

    @property
    def is_hold(self) -> bool:
        """
        Returns:
            Whether the trade type is non-existent (i.e. hold).
        """
        return self == TradeType.HOLD

    @property
    def is_buy(self) -> bool:
        """
        Returns:
            Whether the trade type is a buy offer.
        """
        return self == TradeType.MARKET_BUY or self == TradeType.LIMIT_BUY

    @property
    def is_sell(self) -> bool:
        """
        Returns:
            Whether the trade type is a sell offer.
        """
        return self == TradeType.MARKET_SELL or self == TradeType.LIMIT_SELL


# As a user, I'd like to pull the attached process strategy, pull the data, process the data, get a decision from that data
import datetime as dt
import random
import time
import uuid
from typing import Dict, List

from addict import Dict as ADict
from jamboree import Jamboree
from jamboree.handlers.default.multi import MultiDataManagement
from jamboree.handlers.default.search import BaseSearchHandler
from loguru import logger

from see137 import PortfolioHandler, RequirementsHandler
from see137.common import (AddictPipeline, AddictWithManagerComponent,
                                    ComponentAbstract, Pipeline, PipelineA)
from see137.components.backtest import CreateBacktestComponent
from see137.components.portfolio import CreateEpisodePortfolio
from see137.components.requirements import (AttachToEpisode,
                                                     CreateRequirements)
from see137.components.strategy import CheckStrategy
from see137.components.universe import CheckUniverse
from see137.handlers.strategy.core import StrategyEngine
from see137.models.backtests_models import BacktestConfig
from see137.search.prototype import SearchFacade

logger.disable('jamboree')
# logger.disable(__name__)


def create_backtest_config():
    default_backtest = BacktestConfig()
    default_backtest.universe = "92458de9072d46c8be1883745bc24b37"
    default_backtest.strategy = "53970e2f7fae4990978b61c8cd0e5b57"
    backtest_config = ADict()
    backtest_config.config = default_backtest.dict()
    backtest_config.failed = False
    backtest_config.msg = "Currently starting backtest"
    return backtest_config



def main():
    """
        Functional Pipeline
    """
    processor = Jamboree()

    # EpisodeSpecificInformation()

    create_backtest = AddictPipeline(name="CreateBacktest",
                                     processor=processor)
    # Use the
    create_backtest.pipe([
        Pipeline(name="VerifyBacktest",
                 processor=processor,
                 steps=[
                     CheckUniverse(),
                     CheckStrategy(),
                     CreateBacktestComponent()
                 ])
    ])

    episode_attachment = AttachToEpisode()
    episode_attachment.processor = processor
    # Use this pipeline to setup a backtest to be run
    setup_backtest = (Pipeline(name="SetupBacktest",
                               processor=processor,
                               steps=[
                                   CheckUniverse(),
                                   CheckStrategy(),
                                   CreateEpisodePortfolio(),
                                   CreateRequirements(),
                                   episode_attachment
                               ]))

    backtest_config = create_backtest_config()
    created_backtest = create_backtest.step(backtest_config,
                                            # override=True,
                                            is_loud=False)
    logger.info(created_backtest)
    ran_backtest = setup_backtest.step(created_backtest, is_loud=False)
    strategyid = episode_attachment.spec.strategy_by_episode(ran_backtest.current_episode)
    strat_engine = StrategyEngine(processor=processor)
    current_strategy_engine = strat_engine.pick(strategyid)
    current_strategy = current_strategy_engine.strategy
    logger.critical(current_strategy)


if __name__ == "__main__":
    main()



from .utils.parameters import ModelParameters, is_acceptable_type
from .utils.parameters import convert_to_prices, convert_to_returns, brownian_motion_log_returns
from .main import generate_super_price


from abc import ABC
from typing import Any

class StrategyAbstract(ABC):
    def __init__(self):
        self._data = None
        self._processed = None
    
    @property
    def preprocessed(self):
        if self._processed is None:
            raise AttributeError("Preprocessed data hasn't been set")
        return self._processed
    
    @property
    def information(self):
        if self._data is None:
            raise AttributeError("Data hasn't been set yet")
        return self._data
    
    @information.setter
    def information(self, _data:Any):
        self._data = _data

    def data_check(self):
        raise NotImplementedError("Data check isn't there")

    def process_data(self) -> Any:
        """ Store preprocess the data and ensure it's stored here """
        raise NotImplementedError

    def get_decision(self) -> Any:
        """ Get the preprocessed data and ensure"""
        raise NotImplementedError

    def step(self, data:Any):
        """ Should return an action tuple. 
            
            parameter:
                data: Any

            return:
                action: tuple
                    ([0 - 2], [0 - 99]) 
        """
        self._data = data
        self.data_check()
        self.process_data()
        return self.get_decision()
        
    
    def reset(self):
        self._data = None
        self._processed = None

class ExampleStrategy(StrategyAbstract):
    def process_data(self):
        self.information = self.information + 1
    
    def data_check(self):
        pass

    def get_decision(self):

        if self.information == 1:
            return 0
        elif self.information == 2:
            return 1
        else:
            return 2


"""Separate search for the universe of assets. 
"""
from jamboree.handlers.default.search import BaseSearchHandler
class UniverseMetadata(BaseSearchHandler):
    def __init__(self):
        self.entity = "metadata"
        self.required = {
            "name": str,
            "category": str,
            "metatype": str,
            "submetatype": str,
            "abbreviation": str,
            "subcategories": dict
        }
    
    @property
    def search(self):
        self.reset()
        self['metatype'] = "multi_data_management"
        self['category'] = "universe"
        self["submetatype"] = "price_bag"
        return self



import uuid
from typing import Any, Dict

from addict import Dict as ADict
from loguru import logger

from see137.common import AddictWithManagerComponent

def uhex():
    return uuid.uuid4().hex

class PullStrategyComponent(AddictWithManagerComponent):
    """Find Strategy

        Given a strategy id, pull the strategy object.
    """
    def __init__(self):
        super().__init__()
        self.name = "Find Strategy"
        # Backtest configuration dict. Can be saved into the database immediately.
        self.required_items = {"strategy": str}

    def process(self, item: ADict):
        logger.critical(item)
        return item

    def step(self, item: ADict, **kwargs):
        return super().step(item, **kwargs)


if __name__ == "__main__":
    sample_backtest_dict = ADict(strategy=uhex())
    find_strategy = PullStrategyComponent()
    find_strategy.step(sample_backtest_dict, info={"one": "two"})
